
# This is a relay I run on a personal server. If it gets too expensive to
# run, I'll shut it down.
RENDEZVOUS_RELAY = "http://wormhole-relay.petmail.org:3000/wormhole-relay/"
TRANSIT_RELAY = "tcp:wormhole-transit-relay.petmail.org:3001"
